/********************************************************************
 * FILE: ama.c
 * AUTHOR: Fabian Buske
 * CREATE DATE: 03/07/2008
 * PROJECT: MEME suite
 * COPYRIGHT: 2008, UQ
 *
 * AMA (average motif affinity) is part of an implementation of the
 * algorithm described in
 * "Associating transcription factor binding site motifs with target
 * Go terms and target genes"
 * authors: Mikael Boden and Timothy L. Bailey
 * published: Nucl. Acids Res (2008)
 *
 * The implementation is based on the fimo class in order to
 * use the scoring schemes described in the publication.
 *
 * AMA works on DNA data only.
 ********************************************************************/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "matrix.h"
#include "alphabet.h"
#include "cisml.h"
#include "fasta-io.h"
#include "meme-io.h"
#include "string-list.h"
#include "simple-getopt.h"
#include "zscore.h"
#include "read_tamo.h"
#include "array.h"
#include "alphabet.h"
#include "macros.h"
#include "motif.h"
#include "matrix.h"

#include "ama_scan.h"

const int GFF_FORMAT = 0; // GFF file format
const int CISML_FORMAT = 1; // cisML file format

const int MEME_FORMAT = 1; // input format is meme
const int TAMO_FORMAT = 2; // input format is tamo

const int SEQUENCE_BG = 1; // background frequencies taken from sequence file
const int MOTIF_BG = 2; // background frequencies taken from motif file
const int FILE_BG = 3; // background frequencies taken from specified file
char* program_name = NULL;

// Nucleotide alphabet order as in motif.h
extern char alphabet[];

// As from our ama.h - this is done this way
// to avoid linking issues
VERBOSE_T verbosity = NORMAL_VERBOSE;

/**********************************************************************
 print_score()

 outputs the scores in a gff format
 **********************************************************************/
void print_score(char* program_name, CISML_T* cisml, FILE* gff_file) {

	// iterate over all patterns
	int num_patterns = get_cisml_num_patterns(cisml);
	if (num_patterns > 0) {
		PATTERN_T **patterns = get_cisml_patterns(cisml);
		int i = 0;
		for (i = 0; i < num_patterns; ++i) {
			// iterate over all sequences
			int num_seq = 0;
			num_seq = get_pattern_num_scanned_sequences(patterns[i]);
			SCANNED_SEQUENCE_T **sequences = get_pattern_scanned_sequences(
					patterns[i]);
			int j = 0;
			for (j = 0; j < num_seq; ++j) {
				double score = 0.0;

				if (has_scanned_sequence_pvalue(sequences[j])) {
					score = get_scanned_sequence_score(sequences[j]);
				}
				fprintf(gff_file, "%s", get_scanned_sequence_accession(
						sequences[j]));
				fprintf(gff_file, "\t%s", program_name);
				fprintf(gff_file, "\tsequence");
				fprintf(gff_file, "\t1"); // Start
				fprintf(gff_file, "\t%d", get_scanned_sequence_length(
						sequences[j])); // End
				fprintf(gff_file, "\t%g", score); // Score
				fprintf(gff_file, "\t."); // Strand
				fprintf(gff_file, "\t."); // Frame
				fprintf(gff_file, "\t.\n"); // Comment
			} // num_seq
		} // num_pattern
	} // pattern > 0
}

/*************************************************************************
 * Entry point for ama
 *************************************************************************/
int main(int argc, char *argv[]) {
	char* bg_filename = NULL;
	int max_seq_length = MAX_SEQ;
	char* seq_name = NULL; // Use this sequence name in the output.
	STRING_LIST_T* selected_motifs = NULL;
	double pseudocount = 0.01;
	int output_format = CISML_FORMAT;
	program_name = "ama";
	int scoring = AVG_ODDS;
	int zscoring = 0;
	int input_format = MEME_FORMAT;
	int background = SEQUENCE_BG;
	int num_of_seqs = -1;
	char* print_bg = NULL;
	BOOLEAN_T balance_bf = TRUE;
	BOOLEAN_T status = TRUE;
	BOOLEAN_T scan_both_strands = TRUE;
	clock_t c0, c1; /* measuring cpu_time */
	CISML_T *cisml;
	int i;
	/**********************************************
	 * COMMAND LINE PROCESSING
	 **********************************************/

	const int num_options = 15;
	cmdoption const motif_scan_options[] = {
		{ "bgfile", REQUIRED_VALUE },
		{"bg", REQUIRED_VALUE },
		{ "i-format", REQUIRED_VALUE },
		{ "max-seq-length", REQUIRED_VALUE },
		{ "motif", REQUIRED_VALUE },
		{ "motif-pseudo", REQUIRED_VALUE },
		{ "qvalue", NO_VALUE },
		{ "nobalance", NO_VALUE },
		{ "nostatus", NO_VALUE },
		{ "norc", NO_VALUE },
		{ "o-format", REQUIRED_VALUE },
		{ "print-bg", REQUIRED_VALUE },
		{ "scoring", REQUIRED_VALUE },
		{ "verbosity", REQUIRED_VALUE },
		{ "z-scoring", REQUIRED_VALUE }
	};

	int option_index = 0;

	// Define the usage message.
	char usage[] = "USAGE: ama [options] <motif file> <sequence file>\n"
		"\n"
		"   Options:\n"
		"     --bg [1|2|3] source used to determine background frequencies\n"
		"       (1 = sequence file (default), 2 = meme motif file,\n"
		"       3 = specified background file)\n"
		"     --bgfile <bfile> Read background frequencies from <bfile>.\n"
		"       The default is to obtain frequencies from the specified\n"
		"       <sequence file>.\n"
		"     --max-seq-length <int> Set the maximum length allowed for \n"
		"       input sequences. By default the maximum allowed length \n"
		"       is 250000000\n"
		"     --motif <id> Use only the motif identified by <id>.\n "
		"       This option may be repeated.\n"
		"     --motif-pseudo <float> A pseudocount to be added to each count \n"
		"       in the motif matrix, after first multiplying by the \n"
		"       corresponding background frequency. (default=0.01)\n"
		"     --nobalance Disables the averaging of complementary nucleotide \n"
		"       background frequencies (AT and CG).\n"
		"     --norc Disables the scanning of the reverse complement strand.\n"
		"     --nostatus Suppresses the process information.\n"
		"     --o-format [gff|cisml] output file format (default cisml)\n"
		"     --print-bg <destination file> Prints the background frequencies \n"
		"       to the specified file\n"
		"     --verbosity [1|2|3|4] (default 2)\n"
		"     --scoring [avg-odds|max-odds] Indicates whether the average or \n"
		"       the maximum odds should be calculated (default avg-odds)\n"
		"     --z-scoring <int> Computes the z-score using the specified number \n"
		"       of shuffles. By default z-scoring is disabled (0).\n"
		"\n";

	// Parse the command line.
	if (simple_setopt(argc, argv, num_options, motif_scan_options) != NO_ERROR) {
		die("Error processing command line options: option name too long.\n");
	}

	while (TRUE) {
		int c = 0;
		char* option_name = NULL;
		char* option_value = NULL;
		const char * message = NULL;

		// Read the next option, and break if we're done.
		c = simple_getopt(&option_name, &option_value, &option_index);
		if (c == 0) {
			break;
		} else if (c < 0) {
			(void) simple_getopterror(&message);
			die("Error processing command line options (%s)\n", message);
		}
		if (strcmp(option_name, "bgfile") == 0) {
			bg_filename = option_value;
		} else if (strcmp(option_name, "bg") == 0) {
			if (atoi(option_value) == SEQUENCE_BG)
				background = SEQUENCE_BG;
			else if (atoi(option_value) == MOTIF_BG)
				background = MOTIF_BG;
			else if (atoi(option_value) == FILE_BG)
				background = FILE_BG;
		} else if (strcmp(option_name, "max-seq-length") == 0) {
			max_seq_length = atoi(option_value);
		} else if (strcmp(option_name, "nobalance") == 0) {
			balance_bf = FALSE;
		} else if (strcmp(option_name, "nostatus") == 0) {
			status = FALSE;
		} else if (strcmp(option_name, "norc") == 0) {
			scan_both_strands = FALSE;
		} else if (strcmp(option_name, "motif") == 0) {
			if (selected_motifs == NULL) {
				selected_motifs = new_string_list();
			}
			add_string(option_value, selected_motifs);
		} else if (strcmp(option_name, "motif-pseudo") == 0) {
			pseudocount = atof(option_value);
		} else if (strcmp(option_name, "o-format") == 0) {
			if (strcmp(option_value, "gff") == 0)
				output_format = GFF_FORMAT;
			else if (strcmp(option_value, "cisml") == 0)
				output_format = CISML_FORMAT;
			else {
				if (verbosity >= NORMAL_VERBOSE)
					fprintf(stderr,
							"Output format not known. Use standard instead (cisML)\n");
				output_format = CISML_FORMAT;
			}
		} else if (strcmp(option_name, "verbosity") == 0) {
			verbosity = atoi(option_value);
		} else if (strcmp(option_name, "scoring") == 0) {
			if (strcmp(option_value, "max-odds"))
				scoring = MAX_ODDS;
			else if (strcmp(option_value, "avg-odds"))
				scoring = AVG_ODDS;
			else
				die("Specified scoring scheme not known \n", message);
		} else if (strcmp(option_name, "z-scoring") == 0) {
			zscoring = atoi(option_value);
		}
		// tamo format is not encouraged
		else if (strcmp(option_name, "i-format") == 0) {
			if (strcmp(option_value, "tamo") == 0)
				input_format = TAMO_FORMAT;
			else if (strcmp(option_value, "meme") == 0)
				input_format = MEME_FORMAT;
		} else if (strcmp(option_name, "print-bg") == 0) {
			print_bg = option_value;
		}
	}

	// Must have sequence and motif
	if (argc != option_index + 2) {
		fprintf(stderr, usage);
		exit(EXIT_FAILURE);
	}
	char* motif_filename = argv[option_index];
	option_index++;
	char* fasta_filename = argv[option_index];
	option_index++;

	// measure time
	c0 = clock();

	// FIXME: Alphabet should come from motif file.  WSN, 15 Nov 2007
	set_alphabet(verbosity, "ACGT");

	// Create cisml data structure for recording results
	cisml = allocate_cisml(program_name, motif_filename, fasta_filename);

	/**********************************************
	 * Read the motifs.
	 **********************************************/
	int num_motifs = 0;
	int num_motifs_after_rc = 0;
	MOTIF_T motifs[2 * MAX_MOTIFS];
	STRING_LIST_T* motif_occurrences = NULL;
	BOOLEAN_T has_reverse_strand;
	ARRAY_T* bg_freqs = NULL;
	ARRAY_T* rev_bg_freqs = NULL;
	BOOLEAN_T read_file = FALSE;

	if (input_format == MEME_FORMAT) {
		// Assume MEME XML file.
		read_file = read_meme_xml_file(motif_filename, bg_filename,
				pseudocount, &num_motifs, motifs, &motif_occurrences,
				&has_reverse_strand, &bg_freqs);

		if (read_file == FALSE) {
			// Wasn't MEME XML, try old file format.
			read_meme_file(motif_filename, bg_filename, pseudocount,
					&num_motifs, motifs, &motif_occurrences,
					&has_reverse_strand, &bg_freqs);
		}
	} else if (input_format == TAMO_FORMAT) {
		// read in tamo formatted file
		MOTIF_T* m = NULL;
		read_tamo(motif_filename, pseudocount, &num_motifs, &m);
		int l;
		int max_motifs = min(num_motifs, MAX_MOTIFS);
		for (l = 0; l < max_motifs; ++l) {
			copy_motif(&(m[l]), &(motifs[l]));
		}
		for (l = num_motifs; l > 0; --l) {
			free_motif(&(m[l - 1]));
		}
	}

	// We don't make use of the motif occurrence data
	free_string_list(motif_occurrences);

	// make reverse complements.
	if (scan_both_strands == TRUE) {
		num_motifs_after_rc = num_motifs;
		add_reverse_complements(&num_motifs_after_rc, motifs);
		assert(num_motifs_after_rc == (2 * num_motifs));
	}

	/**************************************************************
	 * Determine number of sequences to scan if not already determined
	 **************************************************************/
	if (num_of_seqs == -1) {
		// Open the FASTA file for reading.
		FILE* fasta_file = NULL;
		if (open_file(fasta_filename, "r", FALSE, "FASTA", "sequences",
				&fasta_file) == 0) {
			die("Couldn't open the file %s.\n", fasta_filename);
		}

		// Read the FASTA file one sequence at a time.
		SEQ_T* sequence = NULL;
		++num_of_seqs;
		while (read_one_fasta(fasta_file, max_seq_length, &sequence)) {
			++num_of_seqs;
		}
	}

	if (verbosity >= NORMAL_VERBOSE) {
		fprintf(stderr, "Number of sequences in file %d.\n", num_of_seqs);
	}

	double workalltogether = num_of_seqs * num_motifs;

	/**************************************************************
	 * Get background frequencies as required
	 **************************************************************/
	if (bg_freqs == NULL) {
		/* Allocate the background array. */
		bg_freqs = allocate_array(get_alph_size(ALL_SIZE));
	}
	if (background == SEQUENCE_BG) {
		// take background from sequence file
		if (verbosity >= NORMAL_VERBOSE) {
			fprintf(stderr,
					"Compute background frequencies from sequence file (%d sequences).\n",
					num_of_seqs);
		}
		// Open the FASTA file for reading.
		FILE* seq_file = NULL;
		if (open_file(fasta_filename, "r", FALSE, "FASTA", "sequences",
				&seq_file) == 0) {
			die("Couldn't open the file %s.\n", fasta_filename);
		}
		// Read the FASTA file one sequence at a time.
		SEQ_T* sequence = NULL;
		double a_freq = 0.0;
		double c_freq = 0.0;
		double g_freq = 0.0;
		double t_freq = 0.0;
		int j, c = 0;
		num_of_seqs = 0;
		while (read_one_fasta(seq_file, max_seq_length, &sequence)) {
			for (j = 0; j < get_seq_length(sequence); ++j) {
				if (toupper(get_seq_char(j, sequence)) == 'A')
					++a_freq;
				else if (toupper(get_seq_char(j, sequence)) == 'C')
					++c_freq;
				else if (toupper(get_seq_char(j, sequence)) == 'G')
					++g_freq;
				else if (toupper(get_seq_char(j, sequence)) == 'T')
					++t_freq;
				++c;
			}
			++num_of_seqs;
		}
		int index = 0;
		index = alphabet_index('A', alphabet);
		set_array_item(index, a_freq / c, bg_freqs);
		index = alphabet_index('C', alphabet);
		set_array_item(index, c_freq / c, bg_freqs);
		index = alphabet_index('G', alphabet);
		set_array_item(index, g_freq / c, bg_freqs);
		index = alphabet_index('T', alphabet);
		set_array_item(index, t_freq / c, bg_freqs);
		fclose(seq_file);

		//	  set_cisml_background_file(cisml, fasta_filename);

		/* Compute values for ambiguous characters. */
		fill_in_ambiguous_chars(FALSE, bg_freqs);
	} else if (background == FILE_BG) {
		// take background from specified file
		if (bg_filename != NULL) {
			bg_freqs = read_background_file(bg_filename);
			set_cisml_background_file(cisml, bg_filename);
		} else {
			fprintf(stderr, "Need to specify a background sequence file!");
			exit(1);
		}
		if (verbosity >= NORMAL_VERBOSE) {
			fprintf(stderr,
					"Take background frequencies from background file (%s).\n",
					bg_filename);
		}
	} else if (background == MOTIF_BG) {
		// take background from motif (already done when meme)
		if (verbosity >= NORMAL_VERBOSE) {
			fprintf(stderr,
					"Take background frequencies from motif file (%s).\n",
					motif_filename);
		}
		if (input_format == TAMO_FORMAT) {
			fprintf(
					stderr,
					"--bg 2 option invalid. The TAMO file format does not contain background frequencies!");
			exit(1);
		}
		//	  set_cisml_background_file(cisml, motif_filename);
	}

	// balance (average) bg frequencies for complementary nucleotides
	if (balance_bf) {
		balance_complementary_dna_freqs(bg_freqs);
	}

	// create reverse complement background frequencies for reverse strand scanning
	if (scan_both_strands) {
		rev_bg_freqs = allocate_array(get_array_length(bg_freqs));
		complement_dna_freqs(bg_freqs, rev_bg_freqs);
	}

	// print background frequencies as requested
	if (print_bg != NULL) {
		FILE *bgFilePtr = fopen(print_bg, "w"); /* Open in write mode */
		fprintf(bgFilePtr, "# Using nucleotide alphabet (%s).\n", alphabet);
		for (i = 0; i < strlen(get_alphabet(FALSE)); ++i) {
			fprintf(bgFilePtr, "%c %g\n", alphabet[i], get_array_item(i,
					bg_freqs));
		}

		fclose(bgFilePtr);
	}

	/**************************************************************
	 * Convert motif matrices into odds matrices
	 **************************************************************/
	int motif_index;
	for (motif_index = 0; motif_index < num_motifs; motif_index++) {
		convert_to_odds_matrix(&(motifs[motif_index]), bg_freqs);
		if (scan_both_strands)
			convert_to_odds_matrix(&(motifs[num_motifs + motif_index]),
					rev_bg_freqs);
	}

	/**************************************************************
	 * Score each of the sites for each of the selected motifs.
	 **************************************************************/
	int counter_motif = 0;

	if (verbosity >= NORMAL_VERBOSE) {
		fprintf(stderr, "Number of motifs in file %d.\n", num_motifs);
	}

	if (status)
		fprintf(stderr, "Processed: 0.0 %% \r");

	/**************************************************************
	 * Read in all sequences
	 **************************************************************/
	// Open the FASTA file for reading.
	FILE* fasta_file = NULL;
	if (open_file(fasta_filename, "r", FALSE, "FASTA", "sequences", &fasta_file)
			== 0) {
		die("Couldn't open the file %s.\n", fasta_filename);
	}

	SEQ_T* sequence = NULL;
	SEQ_T** sequences = NULL;

	int seq_counter = 0;
	int seqchunk = 100;

	while (read_one_fasta(fasta_file, max_seq_length, &sequence)) {
		if ((seq_counter % seqchunk) == 0) {
			Resize(sequences, seq_counter+seqchunk, SEQ_T*);
		}
		sequences[seq_counter++] = sequence;
	}
	/* finish the last entry and line */
	Resize(sequences, seq_counter, SEQ_T*);
	fclose(fasta_file);

	/**************************************************************
	 * Process all motifs
	 **************************************************************/
	for (motif_index = 0; motif_index < num_motifs; motif_index++) {

		MOTIF_T* motif = &(motifs[motif_index]);

		char* motif_id = get_motif_id(motif);
		char* bare_motif_id = motif_id;
		// Drop the strand indicator from the motif id
		if (*bare_motif_id == '+' || *bare_motif_id == '-') {
			bare_motif_id++;
		}

		// Create cisml pattern and add to cisml record
		PATTERN_T *pattern = allocate_pattern(motif_id, "");
		add_cisml_pattern(cisml, pattern);

		if (verbosity >= HIGH_VERBOSE) {
			fprintf(stderr, "Using motif %s of width %d.\n", motif_id,
					motif->length);
		}

		if ((selected_motifs == NULL) || (have_string(bare_motif_id,
				selected_motifs) == TRUE)) {
			// Read the FASTA file one sequence at a time.
			SEQ_T* sequence = NULL;

			int counter_sequence;
			for (counter_sequence = 0; counter_sequence < seq_counter; ++counter_sequence) {
				// check if sequence has sufficient length for the motif
				if (get_seq_length(sequences[counter_sequence]) < motif->length) {
					fprintf(stderr, "%s too short for motif %s. Skip entry!\n",
							get_seq_name(sequences[counter_sequence]), motif_id);
					continue;
				}
				sequence = copy_sequence(sequences[counter_sequence]);
				char* seq_name = get_seq_name(sequence);

				// Create a scanned_sequence record and save it in the pattern.
				SCANNED_SEQUENCE_T *scanned_seq = allocate_scanned_sequence(
						seq_name, seq_name, pattern);
				set_scanned_sequence_length(scanned_seq, get_seq_length(
						sequence));

				MOTIF_T* rev_motif = NULL;
				if (scan_both_strands) {
					rev_motif = &(motifs[num_motifs + motif_index]);
				}

				// scan the sequence using average/maximum motif affinity
				ama_sequence_scan(sequence, motif, rev_motif, scoring,
						zscoring, scanned_seq, scan_both_strands);

				free_seq(sequence);
				sequence = NULL;

				if (status && (counter_sequence % 50) == 0)
					fprintf(stderr, "Processed: %3.1f %% \r", (num_of_seqs
							* counter_motif + counter_sequence) * 100
							/ (workalltogether));
			} // All sequences parsed

			//    	// If q-value option "on" replace p-values by q-values in site list.
			//    	if (compute_qvalues) {
			//    		pattern_calculate_qvalues(pattern);
			//		}
			++counter_motif;
		} else {
			if (verbosity >= NORMAL_VERBOSE) {
				fprintf(stderr, "Skipping motif %s.\n", motif_id);
			}
		}
	} // All motifs parsed

	if (status)
		fprintf(stderr, "Finished          \n");

	// output results
	if (output_format == GFF_FORMAT)
		print_score(program_name, cisml, stdout);
	else if (output_format == CISML_FORMAT)
		print_cisml(stdout, cisml, TRUE, NULL, FALSE);
	else
		die("Output format invalid!\n");

	/**************************************
	 * Clean up.
	 **************************************/
	for (motif_index = 0; motif_index < num_motifs; motif_index++) {
		free_motif(&motifs[motif_index]);
		if (scan_both_strands) {
			free_motif(&motifs[num_motifs + motif_index]);
		}
	}
	free_array(bg_freqs);
	free_cisml(cisml);
	free_string_list(selected_motifs);

	for (i = seq_counter; i > 0; --i) {
		free_seq(sequences[i - 1]);
	}
	myfree(sequences);

	// measure time
	if (verbosity >= NORMAL_VERBOSE) { // starting time
		c1 = clock();
		fprintf(stderr, "cycles (CPU);            %ld cycles\n", (long) c1);
		fprintf(stderr, "elapsed CPU time:        %f seconds\n", (float) (c1
				- c0) / CLOCKS_PER_SEC);
	}

	return (0);
}
